#include "CUDAFunctionsBase.h"

//compute sizes of all regions in the shared memory
template<int numSharedBlocks>
__global__ void raComputeRegionsSizeSharedKernel(int* dOutRegSize,
										   int* dRegionIn, uint pitch, 
										   const int numRegions)
{
	extern __shared__ int dLocalSizes[];
	uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    uint y = (blockIdx.y*blockDim.y)+threadIdx.y;  
	int index = x+(y*pitch);
	int localIndex = threadIdx.x + (threadIdx.y*blockDim.x);
	for(int i=0;i<numSharedBlocks;++i) 
	{	
		if(localIndex < numRegions) dLocalSizes[localIndex] = 0;
		localIndex += (blockIdx.x*blockIdx.y);
	}
	__syncthreads();
	int regIndex = dRegionIn[index];
	if(regIndex >= 0) {
		atomicAdd(dLocalSizes+regIndex, 1);		
	}
	__syncthreads();
	localIndex = threadIdx.x + (threadIdx.y*blockDim.x);
	for(int i=0;i<numSharedBlocks;++i) 
	{	
		if(localIndex < numRegions) {
			int localSize = dLocalSizes[localIndex];
			if(localSize > 0) {
				atomicAdd(dOutRegSize+localIndex, localSize);
			}
		}
		localIndex += (blockIdx.x*blockIdx.y);
	}
}

//compute sizeof of all regions in global memory only (happens when there is a large number of regions)
__global__ void raComputeRegionsSizeKernel(int* dOutRegSize,
										   int* dRegionIn, uint pitch)
{
	uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    uint y = (blockIdx.y*blockDim.y)+threadIdx.y;  
	int index = x+(y*pitch);
	int regIndex = dRegionIn[index];
	if(regIndex >= 0) {
		atomicAdd(dOutRegSize+regIndex, 1);
	}
}

//compute the centroid of each region in shared memory
template<int numSharedBlocks>
__global__ void raComputeRegionsCentroidSharedKernel(int2* dOutRegsCentroid,
												     int* dRegionIn, uint pitch, 												
												     const int numRegions)
{
	extern __shared__ int2 dLocalData[];	
	int* dLocalIntData = (int*)dLocalData;
	int* dOutIntData = (int*)dOutRegsCentroid;
	int x = (blockIdx.x*blockDim.x)+threadIdx.x;
    int y = (blockIdx.y*blockDim.y)+threadIdx.y;  
	int index = x+(y*pitch);
	int localIndex = threadIdx.x + (threadIdx.y*blockDim.x);
	for(int i=0;i<numSharedBlocks;++i) 
	{	
		if(localIndex < numRegions) dLocalData[localIndex] = make_int2(0,0);
		localIndex += (blockIdx.x*blockIdx.y);
	}
	__syncthreads();
	int regIndex = dRegionIn[index];
	if(regIndex >= 0) {		
		regIndex = (regIndex << 1);
		atomicAdd(dLocalIntData+regIndex, x);		
		atomicAdd(dLocalIntData+regIndex+1, y);	
	}
	__syncthreads();
	localIndex = threadIdx.x + (threadIdx.y*blockDim.x);
	for(int i=0;i<numSharedBlocks;++i) 
	{	
		if(localIndex < numRegions) {
			int2 localD = dLocalData[localIndex];
			if(localD.x > 0 || localD.y > 0) {						
				int globalIndex = (localIndex << 1);
				//the results will contain sum of all x's and y's from all elements belonging to the given region
				atomicAdd(dOutIntData+globalIndex, localD.x);
				atomicAdd(dOutIntData+globalIndex+1, localD.y);
			}
		}
		localIndex += (blockIdx.x*blockIdx.y);
	}
}

//compute the centroid of each region only in the global memory
__global__ void raComputeRegionsCentroidGlobalKernel(int2* dOutRegsCentroidSize,
										   int* dRegionIn, uint pitch)
{
	int x = (blockIdx.x*blockDim.x)+threadIdx.x;
    int y = (blockIdx.y*blockDim.y)+threadIdx.y;  
	int index = x+(y*pitch);
	int regIndex = dRegionIn[index];
	if(regIndex >= 0) {
		//the results will contain sum of all x's and y's from all elements belonging to the given region
		atomicAdd((int*)(dOutRegsCentroidSize+regIndex), x);		
		atomicAdd(((int*)(dOutRegsCentroidSize+regIndex))+1, y);								
	}
}

//compute the segment type of each region
__global__ void raComputeRegionTypeKernel(uchar* dOutRegType,
										  int* dRegionIn, uint pitch,
										  uchar* dSegmentedDataIn, uint segPitch)
{
	int x = (blockIdx.x*blockDim.x)+threadIdx.x;
    int y = (blockIdx.y*blockDim.y)+threadIdx.y;  
	int index = x+(y*pitch);
	int regIndex = dRegionIn[index];
	if(regIndex >= 0) {
		uchar type = dSegmentedDataIn[x+(y*segPitch)];
		dOutRegType[regIndex] = type;		//all threads write the same value here so no sync needed
	}
}
