#include "cudaResource.h"
#include <GL/glew.h>
#include <cuda_runtime_api.h>
#include <cuda_gl_interop.h>

CudaResource::CudaResource()
{
	m_resource = 0;
	m_mappedBuffer = 0;
	m_mappedArray = 0;
}

CudaResource::~CudaResource()
{
	Unregister();	
}

void CudaResource::ClearMappedData()
{
	delete m_mappedBuffer;
	delete m_mappedArray;
	m_mappedBuffer = 0;
	m_mappedArray = 0;
}

bool CudaResource::Unregister()
{
	if(m_resource != 0) {
		cudaError_t err = cudaGraphicsUnregisterResource(m_resource);
		if(err != cudaSuccess)
			return false;
	}
	m_resource = 0;
	ClearMappedData();
	return true;
}

bool CudaResource::RegisterTexture(GLenum texTarget, GLuint texId, int w, int h, int d, eCudaResourceMapFlags mapFlags)
{
	Unregister();
	unsigned int flags;
	switch(mapFlags) {
		case eCUDA_RESOURCE_MAP_FLAGS_NONE: flags = cudaGraphicsMapFlagsNone; break;
		case eCUDA_RESOURCE_MAP_FLAGS_WRITE_DISCARD: flags = cudaGraphicsMapFlagsWriteDiscard; break;
		case eCUDA_RESOURCE_MAP_FLAGS_READ_ONLY: flags = cudaGraphicsMapFlagsReadOnly; break;
		default:
			return false;
	}
	cudaError_t err = cudaGraphicsGLRegisterImage(&m_resource, texId, texTarget, flags);	
	if(err != cudaSuccess)
		return false;
	m_texTarget = texTarget;
	m_texDim[0] = w;
	m_texDim[1] = h;
	m_texDim[2] = d;
	return true;
}

bool CudaResource::ChangeMapFlags(eCudaResourceMapFlags mapFlags)
{
	unsigned int flags;
	switch(mapFlags) {
		case eCUDA_RESOURCE_MAP_FLAGS_NONE: flags = cudaGraphicsMapFlagsNone; break;
		case eCUDA_RESOURCE_MAP_FLAGS_WRITE_DISCARD: flags = cudaGraphicsMapFlagsWriteDiscard; break;
		case eCUDA_RESOURCE_MAP_FLAGS_READ_ONLY: flags = cudaGraphicsMapFlagsReadOnly; break;
		default:
			return false;
	}
	cudaError_t err = cudaGraphicsResourceSetMapFlags(m_resource, flags);
	if(err != cudaSuccess)
		return false;	
	return true;
}

bool CudaResource::MapResource()
{	
	cudaError_t err = cudaGraphicsMapResources(1, &m_resource, 0);
	if(err != cudaSuccess)
		return false;
	ClearMappedData();
	return true;
}

bool CudaResource::UnmapResource()
{
	cudaError_t err = cudaGraphicsUnmapResources(1, &m_resource, 0);
	if(err != cudaSuccess)
		return false;
	ClearMappedData();
	return true;
}

CudaArray* CudaResource::GetMappedArray()
{
	if(m_mappedArray) return m_mappedArray;
	cudaArray* array;
	cudaError_t err = cudaGraphicsSubResourceGetMappedArray(&array, m_resource, 0, 0);
	if(err != cudaSuccess)
		return 0;
	m_mappedArray = new CudaArray();
	m_mappedArray->CreateFromCUDAData(array, m_texDim[0], m_texDim[1], m_texDim[2], m_texTarget == GL_TEXTURE_3D);
	return m_mappedArray;
}