#include "regionAnalyzer.h"
#include "cudaBasicUtils.h"
#include "cuda/regionAnalyzerFunctions.h"

RegionAnalyzer::RegionAnalyzer()
{
	m_maxRegions = 0;
	m_threadsX = 16;
	m_threadsY = 16;
	m_imgW = m_imgH = 0;
	m_dataW = m_dataH = 0;
	m_log2DataW = 0;

	m_regionTypeTexId = 0;	
}

RegionAnalyzer::~RegionAnalyzer()
{
	if(m_regionTypeTexId) glDeleteTextures(1, &m_regionTypeTexId);
}

bool RegionAnalyzer::Init(int numRegions, int dataW, int dataH)
{
	//resize internal data based on the number of detected regions
	if(m_maxRegions < numRegions) {
		m_maxRegions = CudaBasicUtils::RoundUp(32, numRegions);
		m_regionSize.Create(sizeof(int)*m_maxRegions);
		m_regionSizeHost.Create(sizeof(int)*m_maxRegions, false);
		m_regionCentroid.Create(sizeof(int)*2*m_maxRegions);
		m_regionCentroidHost.Create(sizeof(int)*2*m_maxRegions, false);
		m_regionType.Create(sizeof(unsigned char)*m_maxRegions);
		m_regionTypeHost.Create(sizeof(unsigned char)*m_maxRegions, false);

		m_int1HostBuf.Create(sizeof(int), false);
		if(!InitTextures(m_maxRegions)) 
			return false;
	}
	dataW = CudaBasicUtils::FirstPow2(dataW);
	dataH = CudaBasicUtils::FirstPow2(dataH);

	m_dataW = dataW;
	m_dataH = dataH;
	m_log2DataW = CudaBasicUtils::Log2Base(m_dataW);
	return true;
}

bool RegionAnalyzer::AnalyzeRegions(CudaBuffer* regionData, CudaBuffer* segmentationData, int numRegions, int imgW, int imgH, int dataW, int dataH)
{
	if(numRegions == 0) return true;
	if(!Init(numRegions, dataW, dataH))
		return false;	
	m_imgW = CudaBasicUtils::RoundUp(m_threadsX, imgW);
	m_imgH = CudaBasicUtils::RoundUp(m_threadsY, imgH);

	if(!m_regionSize.SetZeroData())
		return false;
	if(!m_regionCentroid.SetZeroData())
		return false;
	//analyze regions
	if(numRegions < 4*m_threadsX*m_threadsY) {
		//compute the sizes using the shared memory
		if(!raComputeRegionsSizeShared(m_regionSize.GetData(),  sCudaBuffer2D(regionData->GetData(), regionData->GetPitch()), numRegions, m_imgW, m_imgH, m_threadsX, m_threadsY))
			return false;
		//compute the regions centroids in the shared memory
		if(!raComputeRegionsCentroidsShared(m_regionCentroid.GetData(),  sCudaBuffer2D(regionData->GetData(), regionData->GetPitch()), numRegions, m_imgW, m_imgH, m_threadsX, m_threadsY))
			return false;
	} else {
		//compute the sizes on the global memory alone
		if(!raComputeRegionsSizeGlobal(m_regionSize.GetData(), sCudaBuffer2D(regionData->GetData(), regionData->GetPitch()), m_imgW, m_imgH, m_threadsX, m_threadsY))
			return false;
		//compute centroids in the global mem
		if(!raComputeRegionsCentroidsGlobal(m_regionCentroid.GetData(), sCudaBuffer2D(regionData->GetData(), regionData->GetPitch()), m_imgW, m_imgH, m_threadsX, m_threadsY))
			return false;
	}
	//compute each region type
	if(!raComputeRegionType(m_regionType.GetData(), sCudaBuffer2D(regionData->GetData(), regionData->GetPitch()), sCudaBuffer2D(segmentationData->GetData(), segmentationData->GetPitch()), numRegions, m_imgW, m_imgH, m_threadsX, m_threadsY))
		return false;

	//copy data about regions to the host
	if(!m_regionSizeHost.CopyFrom(&m_regionSize, 0, 0, numRegions*sizeof(int)))
		return false;
	if(!m_regionCentroidHost.CopyFrom(&m_regionCentroid, 0, 0, numRegions*2*sizeof(int)))
		return false;
	if(!m_regionTypeHost.CopyFrom(&m_regionType, 0, 0, numRegions*sizeof(unsigned char)))
		return false;

	if(!TransferDataToTextures(numRegions))
		return false;
	return true;
}

bool RegionAnalyzer::BindRegionTypeTexture(int unit)
{
	if(m_regionTypeTexId != 0) {
		glActiveTexture(GL_TEXTURE0+unit);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_regionTypeTexId);
	}
	return true;
}

bool RegionAnalyzer::InitTextures(int numRegions)
{
	//genrate a texture that will contain a type of segment for each region
	if(m_regionTypeTexId != 0) glDeleteTextures(1, &m_regionTypeTexId);
	glGenTextures(1, &m_regionTypeTexId);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_regionTypeTexId);
	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_LUMINANCE8UI_EXT, numRegions, 1, 0, GL_LUMINANCE_INTEGER_EXT, GL_UNSIGNED_INT, 0);		
	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	if(!m_regionTypeRes.RegisterTexture(GL_TEXTURE_RECTANGLE_ARB, m_regionTypeTexId, numRegions, 1, 1, eCUDA_RESOURCE_MAP_FLAGS_WRITE_DISCARD))
		return false;
	return true;
}

bool RegionAnalyzer::TransferDataToTextures(int numRegions)
{
	//transfer data from CUDA to OpenGL
	if(!m_regionTypeRes.MapResource())
		return false;
	CudaArray* cA = m_regionTypeRes.GetMappedArray();
	if(!cA->CopyFrom(&m_regionType, 0, sizeof(unsigned char)*numRegions))
		return false;
	if(!m_regionTypeRes.UnmapResource())
		return false;
	return true;
}

int RegionAnalyzer::GetRegionType(int regId)
{
	return ((unsigned char*)m_regionTypeHost.GetData())[regId];
}

int RegionAnalyzer::GetRegionSize(int regId)
{
	return ((int*)m_regionSizeHost.GetData())[regId];
}

void RegionAnalyzer::GetRegionCentroid(int regId, float& x, float& y)
{
	float size = (float)GetRegionSize(regId);
	x = (float)((int*)m_regionCentroidHost.GetData())[2*regId] / size;
	y = (float)((int*)m_regionCentroidHost.GetData())[2*regId+1] / size;
}