#include "scene.h"
#include "cudaBasicUtils.h"
#include "profiler.h"
#ifdef WIN32
#include <GL/wglew.h>
#endif
#include <GL/glut.h>
#include <math.h>

#ifndef M_PI
#define M_PI 3.1415926
#endif

Scene::Scene(int initScene, int solver, int resId)
{
	m_winW = m_winH = 1024;
	m_fbo = 0;
	m_rb = 0;
	m_rttTexId = 0;
	m_regionsTexId = 0;
	m_rttTexW = 0;
	m_rttTexH = 0;
	m_displayObjectShader = 0;
	m_displaySceneShader = 0;
	m_displaySceneSubWShader = 0;

	m_selectedRegionId = -1;
	m_selectedRegionBlinkingFrequency = 1.0f;
	m_displayedRegionsTransparency = 0.05f;

	m_displayCentroids = true;
	m_displayRegions = true;
	m_displayHelp = true;
	m_displayLabels = true;

	m_selectedSceneId = initScene;	
	int res = 1024;
	if(resId == 1) res = 2048;
	if(resId == 2) res = 4096;
	m_analysisResolutionX = res;
	m_analysisResolutionY = res;
	
	m_ccl = new CCL();		
}

Scene::~Scene()
{
	delete m_fbo;
	delete m_rb;
	if(m_rttTexId) glDeleteTextures(1, &m_rttTexId);
	if(m_regionsTexId) glDeleteTextures(1, &m_regionsTexId);

	delete m_displayObjectShader;
	delete m_displaySceneShader;
	delete m_displaySceneSubWShader;
	delete m_displayColorShader;

	delete m_ccl;
}

bool Scene::Init()
{
	if(!InitGL()) return false;
	if(!InitShaders()) return false;
	if(!InitScene()) return false;
	m_timer.Start();
	g_profiler.SetEnabled(false);
	return true;
}

bool Scene::InitGL() {
	glewInit();
#ifdef WIN32
	wglSwapIntervalEXT(false);
#endif
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);	
	glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	if(!ResizeRttTex())
		return false;
	return true;
}

bool Scene::InitShaders()
{
	//load shaders that are used for scene visualization and results rendering
	m_displayObjectShader = new ShaderCg();
	if(!m_displayObjectShader->LoadFromFile("../common/shaders/display.cg", "../common/shaders/display.cg", "display_object_v", "display_object_f"))
		return false;
	m_displaySceneShader = new ShaderCg();
	if(!m_displaySceneShader->LoadFromFile(0, "../common/shaders/display.cg", 0, "display_scene_f"))
		return false;
	m_displaySceneSubWShader = new ShaderCg();
	if(!m_displaySceneSubWShader->LoadFromFile(0, "../common/shaders/display.cg", 0, "display_scene_subw_f"))
		return false;
	m_displayColorShader = new ShaderCg();
	if(!m_displayColorShader->LoadFromFile(0, "../common/shaders/display.cg", 0, "display_color_f"))
		return false;
	return true;
}

bool Scene::InitScene() {
	return true;
}


bool Scene::ResizeRttTex()
{
	int newW = CudaBasicUtils::FirstPow2(m_analysisResolutionX);
	int newH = CudaBasicUtils::FirstPow2(m_analysisResolutionY);
	if(newW < m_rttTexW) newW = m_rttTexW;
	if(newH < m_rttTexH) newH = m_rttTexH;
	//if the currentt dimensions of the textures are lower than required then we need to resize the textures
	if(newW > m_rttTexW || newH > m_rttTexH) {
		if(m_rttTexId > 0) {
			glDeleteTextures(1, &m_rttTexId);
			delete m_fbo;
			delete m_rb;
		}		
		//create a main texture into which we render the scene 
		//it is a 8-bit integer texture (so it can be shared directly with CUDA)
		glGenTextures(1, &m_rttTexId);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_rttTexId);
		glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA8UI_EXT, newW, newH, 0, GL_RGBA_INTEGER_EXT, GL_UNSIGNED_BYTE, 0);		
		glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		
		//create a framebuffer and renderbuffer and attach the texture to it
		m_fbo = new FramebufferObject();
		m_rb = new Renderbuffer(GL_DEPTH_COMPONENT24, newW, newH);
		m_fbo->AttachTexture(GL_TEXTURE_RECTANGLE_ARB, m_rttTexId, GL_COLOR_ATTACHMENT0_EXT);	

		m_fbo->AttachRenderBuffer(m_rb->GetId(), GL_DEPTH_ATTACHMENT_EXT);		
	
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
		m_rttTexW = newW;
		m_rttTexH = newH;
		glDisable(GL_CULL_FACE);
		m_fbo->Disable();

		//this texture will contain region ids for all elements in the image (-1 if there they do not belong to any region)
		if(m_regionsTexId > 0) {
			glDeleteTextures(1, &m_regionsTexId);		
		}		
		glGenTextures(1, &m_regionsTexId);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_regionsTexId);
		glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_LUMINANCE32I_EXT, newW, newH, 0, GL_LUMINANCE_INTEGER_EXT, GL_INT, 0);		
		glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		//register this texture in CUDA
		if(!m_regionsTexRes.RegisterTexture(GL_TEXTURE_RECTANGLE_ARB, m_regionsTexId, newW, newH, 1, eCUDA_RESOURCE_MAP_FLAGS_WRITE_DISCARD))
			return false;
		
		//also resize data used in the segmentation unit
		if(!m_segmentation.Init(GL_TEXTURE_RECTANGLE_ARB, m_rttTexId, newW, newH))
			return false;

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_rttTexId);
		glActiveTexture(GL_TEXTURE0+1);
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_regionsTexId);
		glActiveTexture(GL_TEXTURE0);
	}
	return true;
}

void Scene::Display() 
{	
	glClearColor(0,0,0,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();	
	//render the main scene into the texture
	RenderScene();	
	//analyze regions in the scene
	AnalyzeScene();
	//display the results
	m_displaySceneShader->Bind();	
	float transparency = m_displayedRegionsTransparency;
	if(!m_displayRegions) transparency = 1.0f;
	m_displaySceneShader->ChangeUniform1f("regionsTransparency", GPULIB_FRAGMENT_SHADER, transparency);
	m_displaySceneShader->ChangeUniform1i("selectedRegionId", GPULIB_FRAGMENT_SHADER, m_selectedRegionId);
	float selectedRegTransparency = GetSelectedRegionTransparency();
	m_displaySceneShader->ChangeUniform1f("selectedRegionTransparency", GPULIB_FRAGMENT_SHADER, selectedRegTransparency);
	m_displaySceneShader->ChangeUniform2f("texDims", GPULIB_FRAGMENT_SHADER, m_analysisResolutionX, m_analysisResolutionY);
	DisplayRttTex(0, 0, m_winW, m_winH);	
	//display help and other labels
	if(m_displayLabels) DisplayLabels();
	if(m_displayCentroids) DisplayCentroids();
	//display a sub-window that contain the original scene without any results
	if(m_displayLabels) DisplaySubWindow(8*m_winW/10, 8*m_winH/10, m_winW/5, m_winH/5);		
	glutSwapBuffers();
}

void Scene::DisplaySubWindow(int winX, int winY, int winW, int winH)
{	
	glScissor(winX, winY, winW, winH);	
	glEnable(GL_SCISSOR_TEST);
	glClearColor(0.2f, 0.2f, 0.2f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		
	glColor3f(0.2f, 0.2f, 0.2f);
	m_displaySceneSubWShader->Bind();
	m_displaySceneSubWShader->ChangeUniform2f("texDims", GPULIB_FRAGMENT_SHADER, m_analysisResolutionX, m_analysisResolutionY);
	DisplayRttTex(winX+1, winY+1, winW-2, winH-2);
	glDisable(GL_SCISSOR_TEST);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);	
}

void Scene::RenderScene()
{		
	//render the scene into the prepared texture
	if(m_rttTexId == 0) {
		//initialize the texture if it is not already created
		if(!ResizeRttTex()) 
			return;
	}
	m_fbo->Bind();
	glViewport(0, 0, m_analysisResolutionX, m_analysisResolutionY);
	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	
	if(m_selectedSceneId == 0) {
		RenderSceneDemo1();
	}
	if(m_selectedSceneId == 1) {
		RenderSceneDemo2();
	}
	if(m_selectedSceneId == 2) {
		RenderSceneDemo3();
	}
	if(m_selectedSceneId == 3) {
		RenderSceneDemo4();
	}	
	// turn back rendering into the back buffer
	m_fbo->Disable();
	glDrawBuffer(GL_BACK);
	glViewport(0, 0, m_winW, m_winH);
}

void Scene::RenderSceneDemo1()
{
	float time = m_timer.GetTime()/1000.0f;
	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();
	gluPerspective(45,(float)m_analysisResolutionX/(float)m_analysisResolutionY, 0.001, 100.0f);

	m_displayObjectShader->Bind();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();	
	glPushMatrix();	
	glTranslatef(0.0f, 0.0f, -2.0f);
	glRotatef(20, 1.0, 0.0f, 0.0f);			
	glPushMatrix();
	glColor3f(1.0f,0.0f,0.0f);		
	glRotatef(0, 0.0f, 1.0f, 0.0f);
	//the main teapot
	glutSolidTeapot(0.3f);
	glPopMatrix();
	//some spheres aroun the teapot
	int spheresCount = 6;		
	float rotationRad = 0.6f;
	for(int i=0;i<spheresCount;++i) {			
		glPushMatrix();
		glRotatef(-(45.0f)*time, 0.0f, 1.0f, 0.0f);
		glPushMatrix();			
		glTranslatef(rotationRad*sin(2.0f*M_PI*(float)i/(float)spheresCount), 0.0f, rotationRad*cos(2.0f*M_PI*(float)i/(float)spheresCount));
	
		glColor3f(0.0f, 1.0f, 0.0f);
		glutSolidSphere(0.14f, 32, 32);
		glPopMatrix();
		glPopMatrix();
	}	
	glPopMatrix();	
}

void Scene::RenderSceneDemo2()
{
	//display a grid of moving spheres
	float time = m_timer.GetTime()/1000.0f;
	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();
	gluOrtho2D(-1.0f, 1.0f, -1.0f, 1.0f);

	m_displayObjectShader->Bind();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();	
	glPushMatrix();	
	glTranslatef(-0.5f, -0.5f,0.0f);
	unsigned char colors[3][3] = { {255,0,0}, {0,255,0}, {0,0,255}};
	int rows = 10;
	int columns = 10;
	float sphereR = 0.05f;
	glColor3f(1.0f,0.0f,0.0f);	
	for(int i=0;i<rows;++i) {
		glPushMatrix();
		glTranslatef(3.0f*sphereR*sin(0.5f*time+i*M_PI),0.0f,0.0f);
		for(int j=0;j<columns;++j) {
			glColor3ubv(colors[(j)%3]);
			glutSolidSphere(sphereR, 26, 26);
			glTranslatef(1.95f*sphereR, 0.0f, 0.0f);
		}
		glPopMatrix();
		glTranslatef(0.0f, 1.95f*sphereR, 0.0f);
	}

	
	glPopMatrix();	
}


void Scene::RenderSceneDemo3()
{
	//display a randomly distributed spheres - high density
	float time = m_timer.GetTime()/1000.0f;
	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();	
	glOrtho(0, m_analysisResolutionX, 0, m_analysisResolutionY, -m_analysisResolutionX, m_analysisResolutionX);

	m_displayObjectShader->Bind();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();	
	glPushMatrix();		
	unsigned char colors[3][3] = { {255,0,0}, {0,255,0}, {0,0,255}};
	int numSpheres = 2500*((m_analysisResolutionX*m_analysisResolutionX)/(1024*1024));	
	float sphereR = 1024.0f*0.01f;	
	glColor3f(1.0f,0.0f,0.0f);	
	srand(m_timer.GetStartTime());
	for(int i=0;i<numSpheres;++i) {	
		glColor3ubv(colors[(i)%3]);
		glPushMatrix();
		int x = rand()%m_analysisResolutionX;
		glTranslatef(rand()%m_analysisResolutionX, rand()%m_analysisResolutionY, 0.0f);			
		glutSolidSphere(sphereR, 26, 26);		
		glPopMatrix();	
	}
	glPopMatrix();	
}

void Scene::RenderSceneDemo4()
{
	//display a randomly distributed spheres - low density
	float time = m_timer.GetTime()/1000.0f;
	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();	
	glOrtho(0, m_analysisResolutionX, 0, m_analysisResolutionY, -m_analysisResolutionX, m_analysisResolutionX);

	m_displayObjectShader->Bind();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();	
	glPushMatrix();		
	unsigned char colors[3][3] = { {255,0,0}, {0,255,0}, {0,0,255}};
	int numSpheres = 250*((m_analysisResolutionX*m_analysisResolutionX)/(1024*1024));	
	float sphereR = 1024.0f*0.01f;	
	glColor3f(1.0f,0.0f,0.0f);	
	srand(m_timer.GetStartTime());
	for(int i=0;i<numSpheres;++i) {	
		glColor3ubv(colors[(i)%3]);
		glPushMatrix();
		int x = rand()%m_analysisResolutionX;
		glTranslatef(rand()%m_analysisResolutionX, rand()%m_analysisResolutionY, 0.0f);			
		glutSolidSphere(sphereR, 26, 26);		
		glPopMatrix();	
	}

	
	glPopMatrix();	
}


bool Scene::AnalyzeScene()
{
	//first perform a segmentation of the rendered image
	if(!m_segmentation.ApplySegmentation(m_analysisResolutionX, m_analysisResolutionY))
		return false;
	//find the connected region in the segmented image
	if(!m_ccl->FindRegions(m_analysisResolutionX, m_analysisResolutionY, m_segmentation.GetSegmentedData()))
		return false;
	//copy these regions to the prepared texture (map/unmap for CUDA access
	if(!m_regionsTexRes.MapResource())
		return false;
	CudaArray* regTexArray = m_regionsTexRes.GetMappedArray();
	bool ret = true;
	ret = regTexArray->CopyFrom(m_ccl->GetConnectedRegionsBuffer());
	if(!m_regionsTexRes.UnmapResource())
		return false;
	//analyze the regions (compute their sizes and centroids)
	if(!m_ra.AnalyzeRegions(m_ccl->GetConnectedRegionsBuffer(), m_segmentation.GetSegmentedData(), m_ccl->GetNumConnectedRegions(), m_ccl->GetImgWidth(), m_ccl->GetImgHeight(), m_ccl->GetDataWidth(), m_ccl->GetDataHeight()))
		return false;
	return ret;
}

void Scene::DisplayRttTex(int winX, int winY, int winW, int winH)
{	
	//display just a simple quad with the texture containing the results and the regions
	bool ret = m_segmentation.BindSegColorTex(3);
	ret |= m_ra.BindRegionTypeTexture(2);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_rttTexId);
	glActiveTexture(GL_TEXTURE0+1);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_regionsTexId);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0,m_winW, 0, m_winH);
	glMatrixMode(GL_MODELVIEW);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glVertex2f(winX, winY);
	glTexCoord2f(1, 0);
	glVertex2f(winX+winW, winY);
	glTexCoord2f(1, 1);
	glVertex2f(winX+winW, winY+winH);
	glTexCoord2f(0, 1);
	glVertex2f(winX, winY+winH);
	glEnd();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);	
}

void Scene::DisplayLabels()
{	
	char buf[2000];
	m_displayColorShader->Bind();
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, m_winW, m_winH, 0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glColor3f(0.9f, 0.9f, 0.9f);
	if(m_selectedRegionId == -1) DisplayText(10, 30, GLUT_BITMAP_HELVETICA_18, "Left click on any region to select it");
	else DisplayText(10, 30, GLUT_BITMAP_HELVETICA_18, "Press 'Space' to resume the animation");
	
	sprintf(buf, "Total number of connected regions: %d", m_ccl->GetNumConnectedRegions());
	DisplayText(m_winW - 250, m_winH - 100, GLUT_BITMAP_HELVETICA_12, buf);	
	if(m_selectedRegionId == -1) glColor3f(0.9f, 0.5f, 0.5f);
	else glColor3f(0.5f, 0.9f, 0.8f);
	if(m_selectedRegionId == -1) sprintf(buf, "Selected region id: none");
	else sprintf(buf, "Selected region id: %d", m_selectedRegionId);	
	DisplayText(m_winW - 250, m_winH - 85, GLUT_BITMAP_HELVETICA_12, buf);	
	if(m_selectedRegionId != -1) {
		glColor3f(0.5f, 0.8f, 0.5f);
		sprintf(buf, "Region type: ");		
		DisplayText(m_winW - 250, m_winH - 70, GLUT_BITMAP_HELVETICA_12, buf);	
		int type = m_ra.GetRegionType(m_selectedRegionId);
		glColor3ub(m_segmentation.GetSegmentationInterval(type-1)->m_color[0]*2,m_segmentation.GetSegmentationInterval(type-1)->m_color[1]*2, m_segmentation.GetSegmentationInterval(type-1)->m_color[2]*2 );
		sprintf(buf, "%d", type);			
		DisplayOffsetedText(0, 0, GLUT_BITMAP_HELVETICA_12, buf);
		glColor3f(0.5f, 0.8f, 0.5f);
		sprintf(buf, "Region size: ");
		DisplayText(m_winW - 250, m_winH - 55, GLUT_BITMAP_HELVETICA_12, buf);	
		glColor3f(0.8f, 0.8f, 0.8f);
		sprintf(buf, "%d", m_ra.GetRegionSize(m_selectedRegionId));			
		DisplayOffsetedText(0, 0, GLUT_BITMAP_HELVETICA_12, buf);
		glColor3f(0.5f, 0.8f, 0.5f);
		sprintf(buf, "Region centroid: ");
		DisplayText(m_winW - 250, m_winH - 40, GLUT_BITMAP_HELVETICA_12, buf);	
		glColor3f(0.8f, 0.8f, 0.8f);
		float cx, cy;
		m_ra.GetRegionCentroid(m_selectedRegionId, cx, cy);
		sprintf(buf, "(%.2f, %.2f)", cx, cy);
		DisplayOffsetedText(0, 0, GLUT_BITMAP_HELVETICA_12, buf);
	}
	//display help
	if(m_displayHelp) {
		glColor3f(0.9f, 0.9f, 0.9f);
		int helpY = m_winH - 135;
		DisplayText(10, helpY, GLUT_BITMAP_HELVETICA_18, "Help");
		helpY += 15;
		DisplayText(10, helpY, GLUT_BITMAP_HELVETICA_12, "'1', '2', '3', '4' - Select Demo scene");
		helpY += 15;
		DisplayText(10, helpY, GLUT_BITMAP_HELVETICA_12, "'r' - Show/Hide regions");
		helpY += 15;
		DisplayText(10, helpY, GLUT_BITMAP_HELVETICA_12, "'c' - Show/Hide centroids");
		helpY += 15;
		DisplayText(10, helpY, GLUT_BITMAP_HELVETICA_12, "'space' - Pause/Resume animation");
		helpY += 15;
		DisplayText(10, helpY, GLUT_BITMAP_HELVETICA_12, "'o' / 'p' - Decrease/Increase segment 1 min value");
		helpY += 15;
		DisplayText(10, helpY, GLUT_BITMAP_HELVETICA_12, "'k' / 'l' - Decrease/Increase segment 2 min value");
		helpY += 15;
		DisplayText(10, helpY, GLUT_BITMAP_HELVETICA_12, "'n' / 'm' - Decrease/Increase segment 3 min value");
		helpY += 15;
		DisplayText(10, helpY, GLUT_BITMAP_HELVETICA_12, "'h' - Show/Hide help");
	}

	//segmentation
	glColor3f(0.9f, 0.9f, 0.9f);
	int segX = 9*m_winW/10 - 270;
	int segY = 30;
	if(m_winW < 650) {
		segX = 10;
		segY = 60;
	}
	DisplayText(segX, segY , GLUT_BITMAP_HELVETICA_18, "Segmentation");
	glColor3f(0.8f, 0.7f, 0.3f);
	DisplayText(segX, segY+20, GLUT_BITMAP_HELVETICA_12, "Type 1: ");
	glColor3f(0.9f, 0.9f, 0.9f);
	DisplayOffsetedText(0, 0, GLUT_BITMAP_HELVETICA_12, "Min RED: ");
	glColor3f(0.8f, 0.8f, 0.8f);
	sprintf(buf,"%d", (int)m_segmentation.GetSegmentationInterval(0)->m_minVal[0]);
	DisplayOffsetedText(0, 0, GLUT_BITMAP_HELVETICA_12, buf);
	glColor3ub(m_segmentation.GetSegmentationInterval(0)->m_minVal[0], 0, 0);
	DisplayQuad(segX + 150, segY+10, segX + 160, segY+20);
	segY += 15;
	glColor3f(0.8f, 0.7f, 0.3f);
	DisplayText(segX, segY+20, GLUT_BITMAP_HELVETICA_12, "Type 2: ");
	glColor3f(0.9f, 0.9f, 0.9f);
	DisplayOffsetedText(0, 0, GLUT_BITMAP_HELVETICA_12, "Min GREEN: ");
	glColor3f(0.8f, 0.8f, 0.8f);
	sprintf(buf,"%d", (int)m_segmentation.GetSegmentationInterval(1)->m_minVal[1]);
	DisplayOffsetedText(0, 0, GLUT_BITMAP_HELVETICA_12, buf);
	glColor3ub(0, m_segmentation.GetSegmentationInterval(1)->m_minVal[1], 0);
	DisplayQuad(segX + 150, segY+10, segX + 160, segY+20);
	segY += 15;
	glColor3f(0.8f, 0.7f, 0.3f);
	DisplayText(segX, segY+20, GLUT_BITMAP_HELVETICA_12, "Type 3: ");
	glColor3f(0.9f, 0.9f, 0.9f);
	DisplayOffsetedText(0, 0, GLUT_BITMAP_HELVETICA_12, "Min BLUE: ");
	glColor3f(0.8f, 0.8f, 0.8f);
	sprintf(buf,"%d", (int)m_segmentation.GetSegmentationInterval(2)->m_minVal[2]);
	DisplayOffsetedText(0, 0, GLUT_BITMAP_HELVETICA_12, buf);
	glColor3ub(0, 0, m_segmentation.GetSegmentationInterval(2)->m_minVal[2]);
	DisplayQuad(segX + 150, segY+10, segX + 160, segY+20);

	glEnable(GL_DEPTH_TEST);
}

void Scene::DisplayQuad(int minX, int minY, int maxX, int maxY)
{
	glBegin(GL_QUADS);
	glVertex2i(minX, minY);
	glVertex2i(minX, maxY);
	glVertex2i(maxX, maxY);
	glVertex2i(maxX, minY);
	glEnd();
}

void Scene::DisplayText(int x, int y, void* font, char* text, bool rightAlign)
{
	if(rightAlign) {
		x -= glutBitmapLength(font, (unsigned char*)text);
	}
	glRasterPos2i(x, y);
	for(unsigned int i=0;i<strlen(text); ++i) glutBitmapCharacter(font, text[i]);
}

void Scene::DisplayOffsetedText(int x, int y, void* font, char* text)
{
	int pos[4];
	glGetIntegerv(GL_CURRENT_RASTER_POSITION, pos);
	glRasterPos2i(pos[0]+x, m_winH-pos[1]+y);
	for(unsigned int i=0;i<strlen(text); ++i) glutBitmapCharacter(font, text[i]);	
}


void Scene::DisplayCentroids()
{
	//display centroids of all detected regions
	glDisable(GL_DEPTH_TEST);
	m_displayColorShader->Bind();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, m_analysisResolutionX, 0, m_analysisResolutionY);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glColor3f(0.9f, 0.9f, 0.9f);
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(2.0f);
	glEnable(GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	for(int i=0;i<m_ccl->GetNumConnectedRegions();++i) {
		DisplayCentroid(i);
	}
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_BLEND);
	glLineWidth(1.0f);
	glEnable(GL_DEPTH_TEST);
}

void Scene::DisplayCentroid(int regId)
{
	float cX, cY;
	float halfLength = 3.0f;
	m_ra.GetRegionCentroid(regId, cX, cY);
	glBegin(GL_LINES);
	glVertex2f(cX-halfLength, cY-halfLength);
	glVertex2f(cX+halfLength, cY+halfLength);
	glVertex2f(cX-halfLength, cY+halfLength);
	glVertex2f(cX+halfLength, cY-halfLength);
	glEnd();
}

bool Scene::SelectRegionOnPos(int x, int y)
{
	x = (m_analysisResolutionX*x)/m_winW;
	y = (m_analysisResolutionY*(m_winH - y))/m_winH;
	m_selectedRegionId = m_ccl->GetRegionIndexOnPos(x, y);
	return (m_selectedRegionId >= 0);
}

float Scene::GetSelectedRegionTransparency()
{
	if(m_selectedRegionId == -1) return 1.0f;
	float time = m_selectedRegionTimer.GetTime()/1000.0f;
	float f = time * m_selectedRegionBlinkingFrequency;
	f = f - floor(f);
	if(f < 0.5f) return 1.0f*f;
	return 1.0f - 1.0f*f;
}

void Scene::Reshape(int w, int h)
{
	m_winW = w;
	m_winH = h;

	ResizeRttTex();	
	glViewport(0, 0, w, h);
}

void Scene::OnMouseButton(int button, int state, int x, int y)
{
	if(button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
		if(!m_timer.IsPaused()) m_timer.Pause();
		if(SelectRegionOnPos(x, y)) {
			m_selectedRegionTimer.Start();
		}
	}
}

void Scene::OnKey(int key)
{
	if(key == ' ') {
		if(m_timer.IsPaused()) {
			m_timer.Resume();
			m_selectedRegionId = -1;
		}
		else m_timer.Pause();
	}
	if(key == 'r') {
		m_displayRegions = !m_displayRegions;
	}
	if(key == 'c') {
		m_displayCentroids = !m_displayCentroids;
	}
	if(key == 'h') {
		m_displayHelp = !m_displayHelp;
	}
	if(key == 'o') {
		if(m_segmentation.GetSegmentationInterval(0)->m_minVal[0] > 1) m_segmentation.GetSegmentationInterval(0)->m_minVal[0]--;						
		m_selectedRegionId = -1;
	}
	if(key == 'p') {
		if(m_segmentation.GetSegmentationInterval(0)->m_minVal[0] < 255) m_segmentation.GetSegmentationInterval(0)->m_minVal[0]++;				
		m_selectedRegionId = -1;
	}
	if(key == 'k') {
		if(m_segmentation.GetSegmentationInterval(1)->m_minVal[1] > 1) m_segmentation.GetSegmentationInterval(1)->m_minVal[1]--;						
		m_selectedRegionId = -1;
	}
	if(key == 'l') {
		if(m_segmentation.GetSegmentationInterval(1)->m_minVal[1] < 255) m_segmentation.GetSegmentationInterval(1)->m_minVal[1]++;				
		m_selectedRegionId = -1;
	}
	if(key == 'n') {
		if(m_segmentation.GetSegmentationInterval(2)->m_minVal[2] > 1) m_segmentation.GetSegmentationInterval(2)->m_minVal[2]--;						
		m_selectedRegionId = -1;
	}
	if(key == 'm') {
		if(m_segmentation.GetSegmentationInterval(2)->m_minVal[2] < 255) m_segmentation.GetSegmentationInterval(2)->m_minVal[2]++;				
		m_selectedRegionId = -1;
	}	
	if(key == '1') {
		if(m_selectedSceneId != 0) {
			m_selectedSceneId = 0;
			m_selectedRegionId = -1;
		}
	}
	if(key == '2') {
		if(m_selectedSceneId != 1) {
			m_selectedSceneId = 1;
			m_selectedRegionId = -1;
		}
	}
	if(key == '3') {
		if(m_selectedSceneId != 2) {
			m_selectedSceneId = 2;
			m_selectedRegionId = -1;
		}
	}
	if(key == '4') {
		if(m_selectedSceneId != 3) {
			m_selectedSceneId = 3;
			m_selectedRegionId = -1;
		}		
	}
}

void Scene::OnSpecialKey(int key)
{
}
