#include "segmentation.h"
#include "cuda/segmentationFunctions.h"

Segmentation::Segmentation() : m_maxSegIntervals(MAX_SEGMENTATION_INTERVALS)
{
	InitDefaultIntervals();
	m_segColorTexId = 0;
	m_allocatedSegs = 0;
	m_dataInvalid = true;
}

Segmentation::~Segmentation()
{
	if(m_segColorTexId != 0) glDeleteTextures(1, &m_segColorTexId);
}

bool Segmentation::Init(GLenum target, GLuint texId, int w, int h)
{
	//register the input texture containg the scene into CUDA
	if(!m_sceneTexRes.RegisterTexture(target, texId, w, h, 1, eCUDA_RESOURCE_MAP_FLAGS_READ_ONLY))
		return false;
	if(!m_segmentedData.Create2D(sizeof(unsigned char)*w, h))
		return false;	
	if(!m_segmentedData.SetZeroData())
		return false;
	return true;
}

bool Segmentation::BindSegColorTex(int unit)
{
	if(m_segColorTexId == 0 || m_segInterval.size() > m_allocatedSegs)
		CreateSegTextures();
	if(m_dataInvalid) UpdateSegTextures();
	glActiveTexture(GL_TEXTURE0+unit);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_segColorTexId);
	return true;
}

bool Segmentation::CreateSegTextures() 
{
	//create a texture that will holt colors for all segment types (used for visualization only)
	if(m_segColorTexId != 0) glDeleteTextures(1, &m_segColorTexId);
	m_allocatedSegs = m_segInterval.size();
	glGenTextures(1, &m_segColorTexId);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_segColorTexId);
	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA8, m_allocatedSegs, 1, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);		
	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	return true;
}

bool Segmentation::UpdateSegTextures()
{
	//update the segment color texture
	unsigned char* cData = new unsigned char[4*m_segInterval.size()];
	for(unsigned int i=0;i<m_segInterval.size();++i) {
		cData[4*i] = m_segInterval[i].m_color[0];
		cData[4*i+1] = m_segInterval[i].m_color[1];
		cData[4*i+2] = m_segInterval[i].m_color[2];
		cData[4*i+3] = 255;
	}
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, m_segColorTexId);
	glTexSubImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, 0, 0, m_segInterval.size(), 1, GL_RGBA, GL_UNSIGNED_BYTE, cData);
	GLenum err = glGetError();
	if(err != GL_NO_ERROR) {
		delete[] cData;
		return false;
	}
	
	m_dataInvalid = false;
	delete[] cData;
	return true;
}

bool Segmentation::InitDefaultIntervals()
{
	//some "meaningfull" default segmentation thresholding intervals
	SegmentationInterval sI;
	sI.SetMin(1,0,0);
	sI.SetMax(255,0,0);
	sI.SetColor(170, 20, 20);
	m_segInterval.push_back(sI);
	sI.SetMin(0,1,0);
	sI.SetMax(0,255,0);
	sI.SetColor(20, 170, 20);
	m_segInterval.push_back(sI);
	sI.SetMin(0,0,1);
	sI.SetMax(0,0,255);
	sI.SetColor(20, 20, 170);
	m_segInterval.push_back(sI);
	/*sI.SetMin(1,0,0);
	sI.SetMax(255,0,0);
	sI.SetColor(120, 20, 20);
	m_segInterval.push_back(sI);
	sI.SetMin(0,1,0);
	sI.SetMax(0,255,0);
	sI.SetColor(20, 120, 20);
	m_segInterval.push_back(sI);
	sI.SetMin(0,0,1);
	sI.SetMax(0,0,255);
	sI.SetColor(20, 20, 120);
	m_segInterval.push_back(sI);*/
	return true;
}

bool Segmentation::ApplySegmentation(int screenW, int screenH)
{
	if(!m_sceneTexRes.MapResource())
		return false;
	CudaArray* imgArray = m_sceneTexRes.GetMappedArray();
	//reset the previous computed segmented data
	/*if(!m_segmentedData.SetZeroData())
		return false;*/
	int numSegments = m_segInterval.size();
	unsigned char* segData = new unsigned char[6*numSegments];
	for(int i=0;i<numSegments;++i) {
		segData[6*i] = m_segInterval[i].m_minVal[0];
		segData[6*i+1] = m_segInterval[i].m_minVal[1];
		segData[6*i+2] = m_segInterval[i].m_minVal[2];
		segData[6*i+3] = m_segInterval[i].m_maxVal[0];
		segData[6*i+4] = m_segInterval[i].m_maxVal[1];
		segData[6*i+5] = m_segInterval[i].m_maxVal[2];
	}	
	bool ret = applySegmentation(sCudaBuffer2D(m_segmentedData.GetData(), m_segmentedData.GetPitch()), 
								 imgArray->GetArray(), 
								 m_segmentedData.GetSize(0)/sizeof(unsigned char), m_segmentedData.GetSize(1),
								 screenW, screenH, numSegments, segData);
								 
	delete[] segData;
	if(!m_sceneTexRes.UnmapResource())
		return false;
	return ret;
}