#include "shaderCg.h"

#include <iostream>
#include <fstream>
#include <assert.h>

using namespace std;

CGcontext ShaderCg::m_context=0;
CGprofile ShaderCg::m_vertexProfile=CG_PROFILE_UNKNOWN;
CGprofile ShaderCg::m_geometryProfile=CG_PROFILE_UNKNOWN;
CGprofile ShaderCg::m_fragmentProfile=CG_PROFILE_UNKNOWN;

ofstream shaderLog("shadersCg.log");

ShaderCg::ShaderCg()
{
	m_vertexHandle=0;
	m_geometryHandle=0;
	m_fragmentHandle=0;
}


ShaderCg::~ShaderCg()
{	
	if(this == m_activeShader) Unbind();
	if(m_vertexHandle!=0)
	{
		cgDestroyProgram(m_vertexHandle);
	}
	if(m_geometryHandle!=0) {
		cgDestroyProgram(m_geometryHandle);
	}
	if(m_fragmentHandle!=0)
	{
		cgDestroyProgram(m_fragmentHandle);
	}
}

void ShaderCg::Unbind()
{
	if(m_vertexHandle) cgGLDisableProfile(m_vertexProfile);
	if(m_geometryHandle) cgGLDisableProfile(m_geometryProfile);
	if(m_fragmentHandle) cgGLDisableProfile(m_fragmentProfile);
	SetActiveShader(NULL);
}

void ShaderCg::Bind()
{
	Shader* activeShader = Shader::GetActiveShader();	
	if(activeShader!=this)	//check if this program is already bound
	{		
		ShaderCg* cgShader = dynamic_cast<ShaderCg*>(activeShader);
		if(cgShader == NULL) {
			Shader::DisableShaders();
		}
		
		if(cgShader != NULL) {			

			//enable/disable only necessary shaders
			if(cgShader->IsVertexShaderActive() != IsVertexShaderActive()) {
				if(m_vertexHandle==0) cgGLDisableProfile(m_vertexProfile);
				else {
					cgGLBindProgram(m_vertexHandle);
					cgGLEnableProfile(m_vertexProfile);
				}
			} else if(m_vertexHandle!=0) {
				cgGLBindProgram(m_vertexHandle);
			}
			if(cgShader->IsGeometryShaderActive() != IsGeometryShaderActive()) {
				if(m_geometryHandle==0) cgGLDisableProfile(m_geometryProfile);
				else {
					cgGLBindProgram(m_geometryHandle);
					cgGLEnableProfile(m_geometryProfile);
				}
			} else if(m_geometryHandle!=0) {
				cgGLBindProgram(m_geometryHandle);
			}
			if(cgShader->IsFragmentShaderActive() != IsFragmentShaderActive()) {
				if(m_fragmentHandle==0) cgGLDisableProfile(m_fragmentProfile);
				else {
					cgGLBindProgram(m_fragmentHandle);
					cgGLEnableProfile(m_fragmentProfile);
				}
			} else if(m_fragmentHandle!=0) {
				cgGLBindProgram(m_fragmentHandle);
			}
		}
		else {
			if(m_vertexHandle!=0) {
				cgGLBindProgram(m_vertexHandle);
				cgGLEnableProfile(m_vertexProfile);
			}
			if(m_geometryHandle!=0) {
				cgGLBindProgram(m_geometryHandle);
				cgGLEnableProfile(m_geometryProfile);
			}
			if(m_fragmentHandle!=0) {
				cgGLBindProgram(m_fragmentHandle);
				cgGLEnableProfile(m_fragmentProfile);
			}
		}
		Shader::SetActiveShader(this);
	}
}

bool ShaderCg::InitContext() {
	if(m_context==0) {
		m_context=cgCreateContext();
		assert(m_context!=0);
		cgGLSetDebugMode( CG_FALSE );
		m_vertexProfile=cgGLGetLatestProfile(CG_GL_VERTEX);
		assert(cgGLIsProfileSupported(m_vertexProfile)!=CG_FALSE);

#ifndef GPULIB_FORCE_DX9
		m_geometryProfile=cgGLGetLatestProfile(CG_GL_GEOMETRY);
		//assert(cgGLIsProfileSupported(m_geometryProfile)!=CG_FALSE);
#endif		
		m_fragmentProfile=cgGLGetLatestProfile(CG_GL_FRAGMENT);
		assert(m_fragmentProfile!=CG_PROFILE_UNKNOWN);
		cgGLSetOptimalOptions(m_fragmentProfile);

	}
	return true;
}

bool ShaderCg::LoadFromText(const char* vertexShader,const char* fragmentShader,char* vertexMain,char* fragmentMain)
{
	InitContext();
	if(vertexShader)
	{
		m_vertexHandle=cgCreateProgram(m_context,CG_SOURCE,vertexShader,m_vertexProfile,vertexMain,NULL);
		const char* listing = cgGetLastListing(m_context);
		if(listing)
			shaderLog<<listing<<endl;
		if(m_vertexHandle!=0) {
			cgGLLoadProgram(m_vertexHandle);
		} else return 0;
	}

	if(fragmentShader)
	{
		shaderLog<<"Creating cg fragment program from text ";
		shaderLog<<endl;
		m_fragmentHandle=cgCreateProgram(m_context,CG_SOURCE,fragmentShader,m_fragmentProfile,fragmentMain,NULL);
		const char* listing = cgGetLastListing(m_context);
		if(listing)
			shaderLog<<listing<<endl;
		if(m_fragmentHandle!=0) {
			cgGLLoadProgram(m_fragmentHandle);			
		} else return 0;
	}
	return 1;
}


bool ShaderCg::LoadFromFile(const char* vertexShaderFile,const char* fragmentShaderFile,char* vertexMain,char* fragmentMain) {
	InitContext();
	if(vertexShaderFile)
	{
		m_vertexHandle=cgCreateProgramFromFile(m_context,CG_SOURCE,vertexShaderFile,m_vertexProfile,vertexMain,NULL);
		const char* listing = cgGetLastListing(m_context);
		if(listing)
			shaderLog<<listing<<endl;
		if(m_vertexHandle!=0) {
			cgGLLoadProgram(m_vertexHandle);
		} else return 0;
	}

	if(fragmentShaderFile)
	{
		shaderLog<<"Creating cg fragment program from file "<<fragmentShaderFile;
		if(fragmentMain) shaderLog<<" entry point="<<fragmentMain;
		shaderLog<<endl;
		m_fragmentHandle=cgCreateProgramFromFile(m_context,CG_SOURCE,fragmentShaderFile,m_fragmentProfile,fragmentMain,NULL);
		const char* listing = cgGetLastListing(m_context);
		if(listing)
			shaderLog<<listing<<endl;
		if(m_fragmentHandle!=0) {
			cgGLLoadProgram(m_fragmentHandle);
			//shaderLog<<"TST="<<(int)cgGetNamedParameter(_fragmentHandle,"windowDims")<<endl;
			//shaderLog<<"FH="<<(int)_fragmentHandle<<endl;
		} else return 0;
	}
	return 1;
}

bool ShaderCg::LoadFromFileWithGS(const char* vertexShaderFile,const char* geometryShaderFile, const char* fragmentShaderFile,char* vertexMain,char* geometryMain,char* fragmentMain) {
	InitContext();
	if(vertexShaderFile)
	{
		m_vertexHandle=cgCreateProgramFromFile(m_context,CG_SOURCE,vertexShaderFile,m_vertexProfile,vertexMain,NULL);
		const char* listing = cgGetLastListing(m_context);
		if(listing)
			shaderLog<<listing<<endl;
		if(m_vertexHandle!=0) {
			cgGLLoadProgram(m_vertexHandle);
		} else return 0;
	}

	if(geometryShaderFile)
	{
		m_geometryHandle=cgCreateProgramFromFile(m_context,CG_SOURCE,geometryShaderFile,m_geometryProfile,geometryMain,NULL);
		const char* listing = cgGetLastListing(m_context);
		if(listing)
			shaderLog<<listing<<endl;
		if(m_geometryHandle!=0) {
			cgGLLoadProgram(m_geometryHandle);
		} else return 0;
	}

	if(fragmentShaderFile)
	{
		shaderLog<<"Creating cg fragment program from file "<<fragmentShaderFile;
		if(fragmentMain) shaderLog<<" entry point="<<fragmentMain;
		shaderLog<<endl;
		m_fragmentHandle=cgCreateProgramFromFile(m_context,CG_SOURCE,fragmentShaderFile,m_fragmentProfile,fragmentMain,NULL);
		const char* listing = cgGetLastListing(m_context);
		if(listing)
			shaderLog<<listing<<endl;
		if(m_fragmentHandle!=0) {
			cgGLLoadProgram(m_fragmentHandle);
			//shaderLog<<"TST="<<(int)cgGetNamedParameter(_fragmentHandle,"windowDims")<<endl;
			//shaderLog<<"FH="<<(int)_fragmentHandle<<endl;
		} else return 0;
	}
	return 1;
}



CGparameter ShaderCg::InitVertexParameter(const std::string& name) {
	CGparameter param = cgGetNamedParameter(m_vertexHandle, name.c_str());
	if(param == 0) return 0;
	m_vertexParamMap[name] = param;
	return param;
}

CGparameter ShaderCg::GetVertexParameter(const std::string& name) {
	CgParamMap::iterator it = m_vertexParamMap.find(name);
	if(it == m_vertexParamMap.end()) {
		CGparameter param = InitVertexParameter(name);
		if(param == 0) {
			cout<<"warining: param \""<<name<<" not found"<<endl;
			return 0;
		}
		return param;
	}
	return it->second;
}


CGparameter ShaderCg::InitFragmentParameter(const std::string& name) {
	CGparameter param = cgGetNamedParameter(m_fragmentHandle, name.c_str());
	if(param == 0) return 0;
	m_fragmentParamMap[name] = param;
	return param;
}

CGparameter ShaderCg::GetFragmentParameter(const std::string& name) {
	CgParamMap::iterator it = m_fragmentParamMap.find(name);
	if(it == m_fragmentParamMap.end()) {
		CGparameter param = InitFragmentParameter(name);
		if(param == 0) {
			cout<<"warining: param \""<<name<<" not found"<<endl;
			return 0;
		}
		return param;
	}
	return it->second;
}

CGparameter ShaderCg::GetParameter(const std::string& name, eShaderType type) {
	switch(type) {
		case GPULIB_VERTEX_SHADER:
			return GetVertexParameter(name);
		case GPULIB_FRAGMENT_SHADER:
			return GetFragmentParameter(name);
		default:
			return 0;
	};
}

bool ShaderCg::SetUniform1i(const std::string& name, eShaderType type, GLint value) {
	cgSetParameter1i(GetParameter(name,type), value);
	return true;
}

bool ShaderCg::SetUniform2i(const std::string& name, eShaderType type, GLint value0, GLint value1)
{
	cgSetParameter2i(GetParameter(name, type), value0, value1);
	return true;
}

bool ShaderCg::SetUniform1f(const std::string& name, eShaderType type, GLfloat value) {
	cgSetParameter1f(GetParameter(name,type), value);
	return true;
}

bool ShaderCg::SetUniform2f(const std::string& name, eShaderType type, GLfloat value0, GLfloat value1) {
	cgSetParameter2f(GetParameter(name,type), value0, value1);
	return true;
}

bool ShaderCg::SetUniform3f(const std::string& name, eShaderType type, GLfloat value0, GLfloat value1, GLfloat value2) {
	cgSetParameter3f(GetParameter(name,type), value0, value1, value2);
	return true;
}

bool ShaderCg::SetUniform4f(const std::string& name, eShaderType type, GLfloat value0, GLfloat value1, GLfloat value2, GLfloat value3) {
	cgSetParameter4f(GetParameter(name,type), value0, value1, value2, value3);
	return true;
}

bool ShaderCg::SetUniform1fArray(const std::string& name, eShaderType type, int count, const float* data)
{
	cgGLSetParameterArray1f(GetParameter(name,type), 0, count, data);
	return true;
}

bool ShaderCg::SetUniform1iArray(const std::string& name, eShaderType type, int count, const int* data)
{
	cgSetParameterValueic(GetParameter(name, type), count, data);
	return true;
}

bool ShaderCg::SetUniform2fArray(const std::string& name, eShaderType type, int count, const float* data)
{
	cgGLSetParameterArray2f(GetParameter(name,type), 0, count, data);
	return true;
}

bool ShaderCg::SetUniform3fArray(const std::string& name, eShaderType type, int count, const float* data)
{
	cgGLSetParameterArray3f(GetParameter(name,type), 0, count, data);
	return true;
}

bool ShaderCg::SetGLMatrix(const std::string& name, eShaderType type,  GLenum matrix, GLenum transform)
{
	cgGLSetStateMatrixParameter(GetParameter(name, type), (CGGLenum)matrix, (CGGLenum)transform);

	return true;
}